@@ -1,5 +1,5 @@
 /* -*- c-basic-offset: 2 -*- */
-/* Copyright(C) 2009-2012 Brazil
+/* Copyright(C) 2009-2013 Brazil
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -22,6 +22,18 @@
 #include <string.h>
 #include <stdlib.h>
 
+#define GRN_GEO_POINT_IN_NORTH_EAST(point) \
+  ((point)->latitude >= 0 && (point)->longitude >= 0)
+#define GRN_GEO_POINT_IN_NORTH_WEST(point) \
+  ((point)->latitude >= 0 && (point)->longitude < 0)
+#define GRN_GEO_POINT_IN_SOUTH_WEST(point) \
+  ((point)->latitude < 0 && (point)->longitude < 0)
+#define GRN_GEO_POINT_IN_SOUTH_EAST(point) \
+  ((point)->latitude < 0 && (point)->longitude >= 0)
+
+#define GRN_GEO_LONGITUDE_IS_WRAPPED(top_left, bottom_right) \
+  ((top_left)->longitude > 0 && (bottom_right)->longitude < 0)
+
 typedef struct {
   grn_id id;
   double d;
@@ -39,11 +51,14 @@ typedef struct {
   grn_obj bottom_right_point_buffer;
   grn_geo_point *top_left;
   grn_geo_point *bottom_right;
+} in_rectangle_data;
+
+typedef struct {
   grn_geo_point min;
   grn_geo_point max;
   int rectangle_common_bit;
   uint8_t rectangle_common_key[sizeof(grn_geo_point)];
-} in_rectangle_data;
+} in_rectangle_area_data;
 
 static int
 compute_diff_bit(uint8_t *geo_key1, uint8_t *geo_key2)
@@ -905,13 +920,15 @@ grn_geo_select_in_circle(grn_ctx *ctx, grn_obj *index,
   case GRN_DB_TEXT :
   case GRN_DB_LONG_TEXT :
     GRN_OBJ_INIT(&point_on_circle_, GRN_BULK, 0, domain);
-    if (grn_obj_cast(ctx, point_on_circle, &point_on_circle_, GRN_FALSE)) { goto exit; }
+    if (grn_obj_cast(ctx, distance, &point_on_circle_, GRN_FALSE)) { goto exit; }
     point_on_circle = &point_on_circle_;
     /* fallthru */
   case GRN_DB_TOKYO_GEO_POINT :
   case GRN_DB_WGS84_GEO_POINT :
-    if (domain != distance->header.domain) { /* todo */ goto exit; }
-    if (!point_on_circle) { point_on_circle = distance; }
+    if (!point_on_circle) {
+      if (domain != distance->header.domain) { /* todo */ goto exit; }
+      point_on_circle = distance;
+    }
     GRN_GEO_POINT_VALUE(point_on_circle,
                         on_circle.latitude, on_circle.longitude);
     d = distance_raw_func(ctx, center, &on_circle);
@@ -996,19 +1013,15 @@ grn_selector_geo_in_rectangle(grn_ctx *ctx, grn_obj *table, grn_obj *index,
   return ctx->rc;
 }
 
-static grn_rc
-in_rectangle_data_prepare(grn_ctx *ctx, grn_obj *index,
-                          grn_obj *top_left_point,
-                          grn_obj *bottom_right_point,
-                          const char *process_name,
-                          in_rectangle_data *data)
+static void
+in_rectangle_data_fill(grn_ctx *ctx, grn_obj *index,
+                       grn_obj *top_left_point,
+                       grn_obj *bottom_right_point,
+                       const char *process_name,
+                       in_rectangle_data *data)
 {
   grn_id domain;
-
-  if (!index) {
-    ERR(GRN_INVALID_ARGUMENT, "%s: index column is missing", process_name);
-    goto exit;
-  }
+  const char *domain_name;
 
   data->pat = grn_ctx_at(ctx, index->header.domain);
   domain = data->pat->header.domain;
@@ -1028,14 +1041,25 @@ in_rectangle_data_prepare(grn_ctx *ctx, grn_obj *index,
         "%s: index table must be "
         "TokyoGeoPoint or WGS84GeoPoint key type table: <%.*s>",
         process_name, name_size, name);
-    goto exit;
+    return;
+  }
+
+  if (domain == GRN_DB_TOKYO_GEO_POINT) {
+    domain_name = "TokyoGeoPoint";
+  } else {
+    domain_name = "WGS84GeoPoint";
   }
 
   if (top_left_point->header.domain != domain) {
     grn_obj_reinit(ctx, &(data->top_left_point_buffer), domain, GRN_BULK);
     if (grn_obj_cast(ctx, top_left_point, &(data->top_left_point_buffer),
                      GRN_FALSE)) {
-      goto exit;
+      ERR(GRN_INVALID_ARGUMENT,
+          "%s: failed to cast to %s: <%.*s>",
+          process_name, domain_name,
+          (int)GRN_TEXT_LEN(top_left_point),
+          GRN_TEXT_VALUE(top_left_point));
+      return;
     }
     top_left_point = &(data->top_left_point_buffer);
   }
@@ -1045,122 +1069,193 @@ in_rectangle_data_prepare(grn_ctx *ctx, grn_obj *index,
     grn_obj_reinit(ctx, &(data->bottom_right_point_buffer), domain, GRN_BULK);
     if (grn_obj_cast(ctx, bottom_right_point, &(data->bottom_right_point_buffer),
                      GRN_FALSE)) {
-      goto exit;
+      ERR(GRN_INVALID_ARGUMENT,
+          "%s: failed to cast to %s: <%.*s>",
+          process_name, domain_name,
+          (int)GRN_TEXT_LEN(bottom_right_point),
+          GRN_TEXT_VALUE(bottom_right_point));
+      return;
     }
     bottom_right_point = &(data->bottom_right_point_buffer);
   }
   data->bottom_right = GRN_GEO_POINT_VALUE_RAW(bottom_right_point);
+}
 
-  {
-    grn_geo_point *top_left, *bottom_right;
+static void
+in_rectangle_data_validate(grn_ctx *ctx,
+                           const char *process_name,
+                           in_rectangle_data *data)
+{
+  grn_geo_point *top_left, *bottom_right;
 
-    top_left = data->top_left;
-    bottom_right = data->bottom_right;
+  top_left = data->top_left;
+  bottom_right = data->bottom_right;
 
-    if (top_left->latitude < 0 || top_left->longitude < 0 ||
-        bottom_right->latitude < 0 || bottom_right->longitude < 0) {
-      ERR(GRN_FUNCTION_NOT_IMPLEMENTED,
-          "%s: negative coordinate is not implemented.", process_name);
-      goto exit;
-    }
-
-    if (top_left->latitude >= GRN_GEO_MAX_LATITUDE) {
-      ERR(GRN_INVALID_ARGUMENT,
-          "%s: top left point's latitude is too big: "
-          "<%d>(max:%d): (%d,%d) (%d,%d)",
-          process_name,
-          GRN_GEO_MAX_LATITUDE, top_left->latitude,
-          top_left->latitude, top_left->longitude,
-          bottom_right->latitude, bottom_right->longitude);
-      goto exit;
-    }
-
-    if (top_left->longitude >= GRN_GEO_MAX_LONGITUDE) {
-      ERR(GRN_INVALID_ARGUMENT,
-          "%s: top left point's longitude is too big: "
-          "<%d>(max:%d): (%d,%d) (%d,%d)",
-          process_name,
-          GRN_GEO_MAX_LONGITUDE, top_left->longitude,
-          top_left->latitude, top_left->longitude,
-          bottom_right->latitude, bottom_right->longitude);
-      goto exit;
-    }
-
-    if (bottom_right->latitude >= GRN_GEO_MAX_LATITUDE) {
-      ERR(GRN_INVALID_ARGUMENT,
-          "%s: bottom right point's latitude is too big: "
-          "<%d>(max:%d): (%d,%d) (%d,%d)",
-          process_name,
-          GRN_GEO_MAX_LATITUDE, bottom_right->latitude,
-          top_left->latitude, top_left->longitude,
-          bottom_right->latitude, bottom_right->longitude);
-      goto exit;
-    }
-
-    if (bottom_right->longitude >= GRN_GEO_MAX_LONGITUDE) {
-      ERR(GRN_INVALID_ARGUMENT,
-          "%s: bottom right point's longitude is too big: "
-          "<%d>(max:%d): (%d,%d) (%d,%d)",
-          process_name,
-          GRN_GEO_MAX_LONGITUDE, bottom_right->longitude,
-          top_left->latitude, top_left->longitude,
-          bottom_right->latitude, bottom_right->longitude);
-      goto exit;
-    }
+  if (top_left->latitude >= GRN_GEO_MAX_LATITUDE) {
+    ERR(GRN_INVALID_ARGUMENT,
+        "%s: top left point's latitude is too big: "
+        "<%d>(max:%d): (%d,%d) (%d,%d)",
+        process_name,
+        GRN_GEO_MAX_LATITUDE, top_left->latitude,
+        top_left->latitude, top_left->longitude,
+        bottom_right->latitude, bottom_right->longitude);
+    return;
   }
 
-  {
-    int latitude_distance, longitude_distance;
-    int diff_bit;
-    grn_geo_point base;
-    grn_geo_point *top_left, *bottom_right;
-    grn_geo_point *geo_point_input;
-    uint8_t geo_key_input[sizeof(grn_geo_point)];
-    uint8_t geo_key_base[sizeof(grn_geo_point)];
-    uint8_t geo_key_top_left[sizeof(grn_geo_point)];
-    uint8_t geo_key_bottom_right[sizeof(grn_geo_point)];
+  if (top_left->latitude <= GRN_GEO_MIN_LATITUDE) {
+    ERR(GRN_INVALID_ARGUMENT,
+        "%s: top left point's latitude is too small: "
+        "<%d>(min:%d): (%d,%d) (%d,%d)",
+        process_name,
+        GRN_GEO_MIN_LATITUDE, top_left->latitude,
+        top_left->latitude, top_left->longitude,
+        bottom_right->latitude, bottom_right->longitude);
+    return;
+  }
 
-    top_left = data->top_left;
-    bottom_right = data->bottom_right;
+  if (top_left->longitude >= GRN_GEO_MAX_LONGITUDE) {
+    ERR(GRN_INVALID_ARGUMENT,
+        "%s: top left point's longitude is too big: "
+        "<%d>(max:%d): (%d,%d) (%d,%d)",
+        process_name,
+        GRN_GEO_MAX_LONGITUDE, top_left->longitude,
+        top_left->latitude, top_left->longitude,
+        bottom_right->latitude, bottom_right->longitude);
+    return;
+  }
 
-    latitude_distance = top_left->latitude - bottom_right->latitude;
-    longitude_distance = bottom_right->longitude - top_left->longitude;
-    if (latitude_distance > longitude_distance) {
-      geo_point_input = bottom_right;
-      base.latitude = bottom_right->latitude;
-      base.longitude = bottom_right->longitude - longitude_distance;
-    } else {
-      geo_point_input = top_left;
-      base.latitude = top_left->latitude - latitude_distance;
-      base.longitude = top_left->longitude;
-    }
-    grn_gton(geo_key_input, geo_point_input, sizeof(grn_geo_point));
-    grn_gton(geo_key_base, &base, sizeof(grn_geo_point));
-    diff_bit = compute_diff_bit(geo_key_input, geo_key_base);
-    compute_min_and_max(&base, diff_bit, &(data->min), &(data->max));
+  if (top_left->longitude <= GRN_GEO_MIN_LONGITUDE) {
+    ERR(GRN_INVALID_ARGUMENT,
+        "%s: top left point's longitude is too small: "
+        "<%d>(min:%d): (%d,%d) (%d,%d)",
+        process_name,
+        GRN_GEO_MIN_LONGITUDE, top_left->longitude,
+        top_left->latitude, top_left->longitude,
+        bottom_right->latitude, bottom_right->longitude);
+    return;
+  }
 
-    grn_gton(geo_key_top_left, top_left, sizeof(grn_geo_point));
-    grn_gton(geo_key_bottom_right, bottom_right, sizeof(grn_geo_point));
-    data->rectangle_common_bit =
-      compute_diff_bit(geo_key_top_left, geo_key_bottom_right) - 1;
-    compute_min_and_max_key(geo_key_top_left, data->rectangle_common_bit + 1,
-                            data->rectangle_common_key, NULL);
+  if (bottom_right->latitude >= GRN_GEO_MAX_LATITUDE) {
+    ERR(GRN_INVALID_ARGUMENT,
+        "%s: bottom right point's latitude is too big: "
+        "<%d>(max:%d): (%d,%d) (%d,%d)",
+        process_name,
+        GRN_GEO_MAX_LATITUDE, bottom_right->latitude,
+        top_left->latitude, top_left->longitude,
+        bottom_right->latitude, bottom_right->longitude);
+    return;
+  }
+
+  if (bottom_right->latitude <= GRN_GEO_MIN_LATITUDE) {
+    ERR(GRN_INVALID_ARGUMENT,
+        "%s: bottom right point's latitude is too small: "
+        "<%d>(min:%d): (%d,%d) (%d,%d)",
+        process_name,
+        GRN_GEO_MIN_LATITUDE, bottom_right->latitude,
+        top_left->latitude, top_left->longitude,
+        bottom_right->latitude, bottom_right->longitude);
+    return;
+  }
+
+  if (bottom_right->longitude >= GRN_GEO_MAX_LONGITUDE) {
+    ERR(GRN_INVALID_ARGUMENT,
+        "%s: bottom right point's longitude is too big: "
+        "<%d>(max:%d): (%d,%d) (%d,%d)",
+        process_name,
+        GRN_GEO_MAX_LONGITUDE, bottom_right->longitude,
+        top_left->latitude, top_left->longitude,
+        bottom_right->latitude, bottom_right->longitude);
+    return;
+  }
+
+  if (bottom_right->longitude <= GRN_GEO_MIN_LONGITUDE) {
+    ERR(GRN_INVALID_ARGUMENT,
+        "%s: bottom right point's longitude is too small: "
+        "<%d>(min:%d): (%d,%d) (%d,%d)",
+        process_name,
+        GRN_GEO_MIN_LONGITUDE, bottom_right->longitude,
+        top_left->latitude, top_left->longitude,
+        bottom_right->latitude, bottom_right->longitude);
+    return;
+  }
+}
+
+static void
+in_rectangle_area_data_compute(grn_ctx *ctx,
+                               grn_geo_point *top_left,
+                               grn_geo_point *bottom_right,
+                               in_rectangle_area_data *data)
+{
+  int latitude_distance, longitude_distance;
+  int diff_bit;
+  grn_geo_point base;
+  grn_geo_point *geo_point_input;
+  uint8_t geo_key_input[sizeof(grn_geo_point)];
+  uint8_t geo_key_base[sizeof(grn_geo_point)];
+  uint8_t geo_key_top_left[sizeof(grn_geo_point)];
+  uint8_t geo_key_bottom_right[sizeof(grn_geo_point)];
+
+  latitude_distance = top_left->latitude - bottom_right->latitude;
+  longitude_distance = bottom_right->longitude - top_left->longitude;
+  if (latitude_distance > longitude_distance) {
+    geo_point_input = bottom_right;
+    base.latitude = bottom_right->latitude;
+    base.longitude = bottom_right->longitude - longitude_distance;
+  } else {
+    geo_point_input = top_left;
+    base.latitude = top_left->latitude - latitude_distance;
+    base.longitude = top_left->longitude;
+  }
+  grn_gton(geo_key_input, geo_point_input, sizeof(grn_geo_point));
+  grn_gton(geo_key_base, &base, sizeof(grn_geo_point));
+  diff_bit = compute_diff_bit(geo_key_input, geo_key_base);
+  compute_min_and_max(&base, diff_bit, &(data->min), &(data->max));
+
+  grn_gton(geo_key_top_left, top_left, sizeof(grn_geo_point));
+  grn_gton(geo_key_bottom_right, bottom_right, sizeof(grn_geo_point));
+  data->rectangle_common_bit =
+    compute_diff_bit(geo_key_top_left, geo_key_bottom_right) - 1;
+  compute_min_and_max_key(geo_key_top_left, data->rectangle_common_bit + 1,
+                          data->rectangle_common_key, NULL);
 
 #ifdef GEO_DEBUG
-    printf("base:         ");
-    grn_p_geo_point(ctx, &base);
-    printf("min:          ");
-    grn_p_geo_point(ctx, &(data->min));
-    printf("max:          ");
-    grn_p_geo_point(ctx, &(data->max));
-    printf("top-left:     ");
-    grn_p_geo_point(ctx, top_left);
-    printf("bottom-right: ");
-    grn_p_geo_point(ctx, bottom_right);
-    printf("rectangle-common-bit:%10d\n", data->rectangle_common_bit);
-    printf("distance(latitude):  %10d\n", latitude_distance);
-    printf("distance(longitude): %10d\n", longitude_distance);
+  printf("base:         ");
+  grn_p_geo_point(ctx, &base);
+  printf("min:          ");
+  grn_p_geo_point(ctx, &(data->min));
+  printf("max:          ");
+  grn_p_geo_point(ctx, &(data->max));
+  printf("top-left:     ");
+  grn_p_geo_point(ctx, top_left);
+  printf("bottom-right: ");
+  grn_p_geo_point(ctx, bottom_right);
+  printf("rectangle-common-bit:%10d\n", data->rectangle_common_bit);
+  printf("distance(latitude):  %10d\n", latitude_distance);
+  printf("distance(longitude): %10d\n", longitude_distance);
 #endif
+}
+
+static grn_rc
+in_rectangle_data_prepare(grn_ctx *ctx, grn_obj *index,
+                          grn_obj *top_left_point,
+                          grn_obj *bottom_right_point,
+                          const char *process_name,
+                          in_rectangle_data *data)
+{
+  if (!index) {
+    ERR(GRN_INVALID_ARGUMENT, "%s: index column is missing", process_name);
+    goto exit;
+  }
+
+  in_rectangle_data_fill(ctx, index, top_left_point, bottom_right_point,
+                         process_name, data);
+  if (ctx->rc != GRN_SUCCESS) {
+    goto exit;
+  }
+
+  in_rectangle_data_validate(ctx, process_name, data);
+  if (ctx->rc != GRN_SUCCESS) {
+    goto exit;
   }
 
 exit :
@@ -1205,6 +1300,152 @@ exit :
   ((((uint8_t *)(a))[((n_bit) / 8)] &\
     (1 << (7 - ((n_bit) % 8)))) >> (1 << (7 - ((n_bit) % 8))))
 
+static grn_bool
+extract_rectangle_in_area(grn_ctx *ctx,
+                          grn_geo_area_type area_type,
+                          const grn_geo_point *top_left,
+                          const grn_geo_point *bottom_right,
+                          grn_geo_point *area_top_left,
+                          grn_geo_point *area_bottom_right)
+{
+  grn_bool out_of_area = GRN_FALSE;
+  grn_bool cover_all_areas = GRN_FALSE;
+
+  if ((GRN_GEO_POINT_IN_NORTH_WEST(top_left) &&
+       GRN_GEO_POINT_IN_SOUTH_EAST(bottom_right)) ||
+      (GRN_GEO_POINT_IN_NORTH_EAST(top_left) &&
+       GRN_GEO_POINT_IN_SOUTH_WEST(bottom_right))) {
+    cover_all_areas = GRN_TRUE;
+  }
+
+  switch (area_type) {
+  case GRN_GEO_AREA_NORTH_EAST :
+    if (cover_all_areas ||
+        GRN_GEO_POINT_IN_NORTH_EAST(top_left) ||
+        GRN_GEO_POINT_IN_NORTH_EAST(bottom_right)) {
+      area_top_left->latitude     = MAX(top_left->latitude,      0);
+      area_bottom_right->latitude = MAX(bottom_right->latitude,  0);
+      if (GRN_GEO_LONGITUDE_IS_WRAPPED(top_left, bottom_right)) {
+        area_top_left->longitude     = top_left->longitude;
+        area_bottom_right->longitude = GRN_GEO_MAX_LONGITUDE;
+      } else {
+        area_top_left->longitude     = MAX(top_left->longitude,     0);
+        area_bottom_right->longitude = MAX(bottom_right->longitude, 0);
+      }
+    } else {
+      out_of_area = GRN_TRUE;
+    }
+    break;
+  case GRN_GEO_AREA_NORTH_WEST :
+    if (cover_all_areas ||
+        GRN_GEO_POINT_IN_NORTH_WEST(top_left) ||
+        GRN_GEO_POINT_IN_NORTH_WEST(bottom_right)) {
+      area_top_left->latitude     = MAX(top_left->latitude,       0);
+      area_bottom_right->latitude = MAX(bottom_right->latitude,   0);
+      if (GRN_GEO_LONGITUDE_IS_WRAPPED(top_left, bottom_right)) {
+        area_top_left->longitude     = GRN_GEO_MIN_LONGITUDE;
+        area_bottom_right->longitude = bottom_right->longitude;
+      } else {
+        area_top_left->longitude     = MIN(top_left->longitude,     -1);
+        area_bottom_right->longitude = MIN(bottom_right->longitude, -1);
+      }
+    } else {
+      out_of_area = GRN_TRUE;
+    }
+    break;
+  case GRN_GEO_AREA_SOUTH_WEST :
+    if (cover_all_areas ||
+        GRN_GEO_POINT_IN_SOUTH_WEST(top_left) ||
+        GRN_GEO_POINT_IN_SOUTH_WEST(bottom_right)) {
+      area_top_left->latitude     = MIN(top_left->latitude,      -1);
+      area_bottom_right->latitude = MIN(bottom_right->latitude,  -1);
+      if (GRN_GEO_LONGITUDE_IS_WRAPPED(top_left, bottom_right)) {
+        area_top_left->longitude     = GRN_GEO_MIN_LONGITUDE;
+        area_bottom_right->longitude = bottom_right->longitude;
+      } else {
+        area_top_left->longitude     = MIN(top_left->longitude,     -1);
+        area_bottom_right->longitude = MIN(bottom_right->longitude, -1);
+      }
+    } else {
+      out_of_area = GRN_TRUE;
+    }
+    break;
+  case GRN_GEO_AREA_SOUTH_EAST :
+    if (cover_all_areas ||
+        GRN_GEO_POINT_IN_SOUTH_EAST(top_left) ||
+        GRN_GEO_POINT_IN_SOUTH_EAST(bottom_right)) {
+      area_top_left->latitude     = MIN(top_left->latitude,      -1);
+      area_bottom_right->latitude = MIN(bottom_right->latitude,  -1);
+      if (GRN_GEO_LONGITUDE_IS_WRAPPED(top_left, bottom_right)) {
+        area_top_left->longitude     = top_left->longitude;
+        area_bottom_right->longitude = GRN_GEO_MAX_LONGITUDE;
+      } else {
+        area_top_left->longitude     = MAX(top_left->longitude,      0);
+        area_bottom_right->longitude = MAX(bottom_right->longitude,  0);
+      }
+    } else {
+      out_of_area = GRN_TRUE;
+    }
+    break;
+  default :
+    out_of_area = GRN_TRUE;
+    break;
+  }
+
+  return out_of_area;
+}
+
+static void
+grn_geo_cursor_area_init(grn_ctx *ctx,
+                         grn_geo_cursor_area *area,
+                         grn_geo_area_type area_type,
+                         const grn_geo_point *top_left,
+                         const grn_geo_point *bottom_right)
+{
+  grn_geo_point area_top_left, area_bottom_right;
+  in_rectangle_area_data data;
+  grn_geo_cursor_entry *entry;
+  grn_bool out_of_area;
+
+  out_of_area = extract_rectangle_in_area(ctx,
+                                          area_type,
+                                          top_left,
+                                          bottom_right,
+                                          &area_top_left,
+                                          &area_bottom_right);
+  if (out_of_area) {
+    area->current_entry = -1;
+    return;
+  }
+
+  area->current_entry = 0;
+  memcpy(&(area->top_left), &area_top_left, sizeof(grn_geo_point));
+  memcpy(&(area->bottom_right), &area_bottom_right, sizeof(grn_geo_point));
+  grn_gton(area->top_left_key, &area_top_left, sizeof(grn_geo_point));
+  grn_gton(area->bottom_right_key, &area_bottom_right, sizeof(grn_geo_point));
+
+  entry = &(area->entries[area->current_entry]);
+  in_rectangle_area_data_compute(ctx,
+                                 &area_top_left,
+                                 &area_bottom_right,
+                                 &data);
+  entry->target_bit = data.rectangle_common_bit;
+  memcpy(entry->key, data.rectangle_common_key, sizeof(grn_geo_point));
+  entry->status_flags =
+    GRN_GEO_CURSOR_ENTRY_STATUS_TOP_INCLUDED |
+    GRN_GEO_CURSOR_ENTRY_STATUS_BOTTOM_INCLUDED |
+    GRN_GEO_CURSOR_ENTRY_STATUS_LEFT_INCLUDED |
+    GRN_GEO_CURSOR_ENTRY_STATUS_RIGHT_INCLUDED;
+  if (data.min.latitude == area_bottom_right.latitude &&
+      data.max.latitude == area_top_left.latitude) {
+    entry->status_flags |= GRN_GEO_CURSOR_ENTRY_STATUS_LATITUDE_INNER;
+  }
+  if (data.min.longitude == area_top_left.longitude &&
+      data.max.longitude == area_bottom_right.longitude) {
+    entry->status_flags |= GRN_GEO_CURSOR_ENTRY_STATUS_LONGITUDE_INNER;
+  }
+}
+
 grn_obj *
 grn_geo_cursor_open_in_rectangle(grn_ctx *ctx,
                                  grn_obj *index,
@@ -1235,32 +1476,21 @@ grn_geo_cursor_open_in_rectangle(grn_ctx *ctx,
   cursor->index = index;
   memcpy(&(cursor->top_left), data.top_left, sizeof(grn_geo_point));
   memcpy(&(cursor->bottom_right), data.bottom_right, sizeof(grn_geo_point));
-  grn_gton(cursor->top_left_key, data.top_left, sizeof(grn_geo_point));
-  grn_gton(cursor->bottom_right_key, data.bottom_right, sizeof(grn_geo_point));
   cursor->pat_cursor = NULL;
   cursor->ii_cursor = NULL;
   cursor->offset = offset;
   cursor->rest = limit;
 
-  cursor->current_entry = 0;
+  cursor->current_area = GRN_GEO_AREA_NORTH_EAST;
   {
-    grn_geo_cursor_entry *entry;
-
-    entry = &(cursor->entries[cursor->current_entry]);
-    entry->target_bit = data.rectangle_common_bit;
-    memcpy(entry->key, data.rectangle_common_key, sizeof(grn_geo_point));
-    entry->status_flags =
-      GRN_GEO_CURSOR_ENTRY_STATUS_TOP_INCLUDED |
-      GRN_GEO_CURSOR_ENTRY_STATUS_BOTTOM_INCLUDED |
-      GRN_GEO_CURSOR_ENTRY_STATUS_LEFT_INCLUDED |
-      GRN_GEO_CURSOR_ENTRY_STATUS_RIGHT_INCLUDED;
-    if (data.min.latitude == data.bottom_right->latitude &&
-        data.max.latitude == data.top_left->latitude) {
-      entry->status_flags |= GRN_GEO_CURSOR_ENTRY_STATUS_LATITUDE_INNER;
-    }
-    if (data.min.longitude == data.top_left->longitude &&
-        data.max.longitude == data.bottom_right->longitude) {
-      entry->status_flags |= GRN_GEO_CURSOR_ENTRY_STATUS_LONGITUDE_INNER;
+    grn_geo_area_type area_type;
+    const grn_geo_point *top_left = &(cursor->top_left);
+    const grn_geo_point *bottom_right = &(cursor->bottom_right);
+    for (area_type = GRN_GEO_AREA_NORTH_EAST;
+         area_type < GRN_GEO_AREA_LAST;
+         area_type++) {
+      grn_geo_cursor_area_init(ctx, &(cursor->areas[area_type]),
+                               area_type, top_left, bottom_right);
     }
   }
   {
@@ -1311,7 +1541,9 @@ grn_geo_cursor_entry_next_push(grn_ctx *ctx,
                                      GRN_CURSOR_PREFIX|GRN_CURSOR_SIZE_BY_BIT);
   if (pat_cursor) {
     if (grn_table_cursor_next(ctx, pat_cursor)) {
-      next_entry = &(cursor->entries[++cursor->current_entry]);
+      grn_geo_cursor_area *area;
+      area = &(cursor->areas[cursor->current_area]);
+      next_entry = &(area->entries[++area->current_entry]);
       memcpy(next_entry, entry, sizeof(grn_geo_cursor_entry));
       pushed = GRN_TRUE;
     }
@@ -1326,16 +1558,28 @@ grn_geo_cursor_entry_next(grn_ctx *ctx,
                           grn_geo_cursor_in_rectangle *cursor,
                           grn_geo_cursor_entry *entry)
 {
-  uint8_t *top_left_key = cursor->top_left_key;
-  uint8_t *bottom_right_key = cursor->bottom_right_key;
+  uint8_t *top_left_key;
+  uint8_t *bottom_right_key;
   int max_target_bit = GRN_GEO_KEY_MAX_BITS - cursor->minimum_reduce_bit;
+  grn_geo_cursor_area *area = NULL;
 
-  if (cursor->current_entry < 0) {
+  while (cursor->current_area < GRN_GEO_AREA_LAST) {
+    area = &(cursor->areas[cursor->current_area]);
+    if (area->current_entry >= 0) {
+      break;
+    }
+    cursor->current_area++;
+    area = NULL;
+  }
+
+  if (!area) {
     return GRN_FALSE;
   }
 
+  top_left_key = area->top_left_key;
+  bottom_right_key = area->bottom_right_key;
   memcpy(entry,
-         &(cursor->entries[cursor->current_entry--]),
+         &(area->entries[area->current_entry--]),
          sizeof(grn_geo_cursor_entry));
   while (GRN_TRUE) {
     grn_geo_cursor_entry next_entry0, next_entry1;
@@ -1514,9 +1758,9 @@ grn_geo_cursor_entry_next(grn_ctx *ctx,
 
       printf("%d: pushed\n", entry->target_bit);
       printf("stack:\n");
-      for (i = cursor->current_entry; i >= 0; i--) {
+      for (i = area->current_entry; i >= 0; i--) {
         grn_geo_cursor_entry *stack_entry;
-        stack_entry = &(cursor->entries[i]);
+        stack_entry = &(area->entries[i]);
         printf("%2d: ", i);
         inspect_key(ctx, stack_entry->key);
         printf("    ");
@@ -1524,7 +1768,7 @@ grn_geo_cursor_entry_next(grn_ctx *ctx,
       }
 #endif
       memcpy(entry,
-             &(cursor->entries[cursor->current_entry--]),
+             &(area->entries[area->current_entry--]),
              sizeof(grn_geo_cursor_entry));
 #ifdef GEO_DEBUG
       printf("%d: pop entry\n", entry->target_bit);
@@ -1771,6 +2015,7 @@ grn_geo_estimate_in_rectangle(grn_ctx *ctx,
     int total_latitude_distance, total_longitude_distance;
     double select_ratio;
     double estimated_n_records;
+    in_rectangle_area_data area_data;
 
     rc = geo_point_get(ctx, data.pat, GRN_CURSOR_ASCENDING, &min);
     if (!rc) {
@@ -1786,8 +2031,14 @@ grn_geo_estimate_in_rectangle(grn_ctx *ctx,
       goto exit;
     }
 
-    select_latitude_distance = abs(data.max.latitude - data.min.latitude);
-    select_longitude_distance = abs(data.max.longitude - data.min.longitude);
+    in_rectangle_area_data_compute(ctx,
+                                   data.top_left,
+                                   data.bottom_right,
+                                   &area_data);
+    select_latitude_distance =
+      abs(area_data.max.latitude - area_data.min.latitude);
+    select_longitude_distance =
+      abs(area_data.max.longitude - area_data.min.longitude);
     total_latitude_distance = abs(max.latitude - min.latitude);
     total_longitude_distance = abs(max.longitude - min.longitude);
 
@@ -1865,8 +2116,8 @@ grn_geo_in_circle(grn_ctx *ctx, grn_obj *point, grn_obj *center,
     case GRN_DB_WGS84_GEO_POINT :
       if (domain != radius_or_point->header.domain) { /* todo */ goto exit; }
       r = d <= distance_raw_func(ctx,
-                          GRN_GEO_POINT_VALUE_RAW(radius_or_point),
-                          GRN_GEO_POINT_VALUE_RAW(center));
+                                 GRN_GEO_POINT_VALUE_RAW(radius_or_point),
+                                 GRN_GEO_POINT_VALUE_RAW(center));
       break;
     default :
       goto exit;
@@ -1882,10 +2133,30 @@ grn_bool
 grn_geo_in_rectangle_raw(grn_ctx *ctx, grn_geo_point *point,
                          grn_geo_point *top_left, grn_geo_point *bottom_right)
 {
-    return ((top_left->longitude <= point->longitude) &&
-            (point->longitude <= bottom_right->longitude) &&
-            (bottom_right->latitude <= point->latitude) &&
-            (point->latitude <= top_left->latitude));
+  if (point->latitude > top_left->latitude) {
+    return GRN_FALSE;
+  }
+  if (point->latitude < bottom_right->latitude) {
+    return GRN_FALSE;
+  }
+
+  if (GRN_GEO_LONGITUDE_IS_WRAPPED(top_left, bottom_right)) {
+    if (point->longitude >= top_left->longitude) {
+      return GRN_TRUE;
+    }
+    if (point->longitude <= bottom_right->longitude) {
+      return GRN_TRUE;
+    }
+    return GRN_FALSE;
+  } else {
+    if (point->longitude < top_left->longitude) {
+      return GRN_FALSE;
+    }
+    if (point->longitude > bottom_right->longitude) {
+      return GRN_FALSE;
+    }
+    return GRN_TRUE;
+  }
 }
 
 grn_bool
