@@ -1,5 +1,5 @@
 /* -*- c-basic-offset: 2 -*- */
-/* Copyright(C) 2009-2013 Brazil
+/* Copyright(C) 2009-2014 Brazil
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -580,7 +580,7 @@ grn_text_atoj(grn_ctx *ctx, grn_obj *outbuf, grn_content_type output_type,
   } else {
     grn_obj_format *format_argument = NULL;
     grn_obj_format format;
-    format.flags = 0;
+    GRN_OBJ_FORMAT_INIT(&format, 0, 0, 0, 0);
     switch (obj->header.type) {
     case GRN_COLUMN_FIX_SIZE :
       GRN_VALUE_FIX_SIZE_INIT(&buf, 0, DB_OBJ(obj)->range);
@@ -593,10 +593,10 @@ grn_text_atoj(grn_ctx *ctx, grn_obj *outbuf, grn_content_type output_type,
           GRN_VALUE_FIX_SIZE_INIT(&buf, GRN_OBJ_VECTOR, DB_OBJ(obj)->range);
         } else {
           GRN_VALUE_VAR_SIZE_INIT(&buf, GRN_OBJ_VECTOR, DB_OBJ(obj)->range);
-          if (obj->header.flags & GRN_OBJ_WITH_WEIGHT) {
-            format.flags |= GRN_OBJ_FORMAT_WITH_WEIGHT;
-            format_argument = &format;
-          }
+        }
+        if (obj->header.flags & GRN_OBJ_WITH_WEIGHT) {
+          format.flags |= GRN_OBJ_FORMAT_WITH_WEIGHT;
+          format_argument = &format;
         }
       } else {
         GRN_VALUE_VAR_SIZE_INIT(&buf, 0, DB_OBJ(obj)->range);
@@ -751,122 +751,182 @@ grn_output_bulk(grn_ctx *ctx, grn_obj *outbuf, grn_content_type output_type,
   GRN_OBJ_FIN(ctx, &buf);
 }
 
+static void
+grn_output_uvector_result_set(grn_ctx *ctx,
+                              grn_obj *outbuf,
+                              grn_content_type output_type,
+                              grn_obj *uvector,
+                              grn_obj_format *format)
+{
+  unsigned int i_hit, n_hits;
+  unsigned int i_column, n_columns;
+  unsigned int n_elements;
+  grn_obj **columns;
+  grn_obj buf;
+  grn_bool with_column_names = GRN_FALSE;
+
+  n_hits = grn_vector_size(ctx, uvector);
+
+  n_columns = GRN_BULK_VSIZE(&format->columns) / sizeof(grn_obj *);
+  columns = (grn_obj **)GRN_BULK_HEAD(&format->columns);
+
+  GRN_TEXT_INIT(&buf, 0);
+
+  if (n_hits > 0 && format->flags & GRN_OBJ_FORMAT_WITH_COLUMN_NAMES) {
+    with_column_names = GRN_TRUE;
+  }
+
+  n_elements = 1; /* for NHITS */
+  if (with_column_names) {
+    n_elements += 1; /* for COLUMNS */
+  }
+  n_elements += n_hits; /* for HITS */
+  grn_output_array_open(ctx, outbuf, output_type, "RESULTSET", n_elements);
+
+  grn_output_array_open(ctx, outbuf, output_type, "NHITS", 1);
+  grn_text_itoa(ctx, outbuf, n_hits);
+  grn_output_array_close(ctx, outbuf, output_type);
+
+  if (with_column_names) {
+    grn_output_array_open(ctx, outbuf, output_type, "COLUMNS", n_columns);
+    for (i_column = 0; i_column < n_columns; i_column++) {
+      grn_id range_id;
+      grn_output_array_open(ctx, outbuf, output_type, "COLUMN", 2);
+
+      /* name */
+      GRN_BULK_REWIND(&buf);
+      grn_column_name_(ctx, columns[i_column], &buf);
+      grn_output_obj(ctx, outbuf, output_type, &buf, NULL);
+
+      /* type */
+      range_id = grn_obj_get_range(ctx, columns[i_column]);
+      if (range_id == GRN_ID_NIL) {
+        GRN_TEXT_PUTS(ctx, outbuf, "null");
+      } else {
+        int name_len;
+        grn_obj *range_obj;
+        char name_buf[GRN_TABLE_MAX_KEY_SIZE];
+
+        range_obj = grn_ctx_at(ctx, range_id);
+        name_len = grn_obj_name(ctx, range_obj, name_buf,
+                                GRN_TABLE_MAX_KEY_SIZE);
+        GRN_BULK_REWIND(&buf);
+        GRN_TEXT_PUT(ctx, &buf, name_buf, name_len);
+        grn_output_obj(ctx, outbuf, output_type, &buf, NULL);
+      }
+
+      grn_output_array_close(ctx, outbuf, output_type);
+    }
+    grn_output_array_close(ctx, outbuf, output_type);
+  }
+
+  for (i_hit = 0; i_hit < n_hits++; i_hit++) {
+    grn_id id;
+
+    id = grn_uvector_get_element(ctx, uvector, i_hit, NULL);
+    grn_output_array_open(ctx, outbuf, output_type, "HITS", n_columns);
+    for (i_column = 0; i_column < n_columns; i_column++) {
+      GRN_BULK_REWIND(&buf);
+      grn_obj_get_value(ctx, columns[i_column], id, &buf);
+      grn_output_obj(ctx, outbuf, output_type, &buf, NULL);
+    }
+    grn_output_array_close(ctx, outbuf, output_type);
+  }
+
+  grn_output_array_close(ctx, outbuf, output_type);
+
+  GRN_OBJ_FIN(ctx, &buf);
+}
+
 static inline void
 grn_output_uvector(grn_ctx *ctx, grn_obj *outbuf, grn_content_type output_type,
                    grn_obj *uvector, grn_obj_format *format)
 {
-  grn_obj buf;
-  GRN_TEXT_INIT(&buf, 0);
+  grn_bool output_result_set = GRN_FALSE;
+  grn_bool with_weight = GRN_FALSE;
+  grn_obj *range;
+  grn_bool range_is_type;
+
   if (format) {
-    int i, j;
-    grn_id *v = (grn_id *)GRN_BULK_HEAD(uvector), *ve = (grn_id *)GRN_BULK_CURR(uvector);
-    int ncolumns = GRN_BULK_VSIZE(&format->columns) / sizeof(grn_obj *);
-    grn_obj **columns = (grn_obj **)GRN_BULK_HEAD(&format->columns);
-    grn_output_array_open(ctx, outbuf, output_type, "RESULTSET", -1);
-    grn_output_array_open(ctx, outbuf, output_type, "NHITS", 1);
-    grn_text_itoa(ctx, outbuf, ve - v);
-    grn_output_array_close(ctx, outbuf, output_type);
-    if (v < ve) {
-      if (format->flags & GRN_OBJ_FORMAT_WITH_COLUMN_NAMES) {
-        grn_output_array_open(ctx, outbuf, output_type, "COLUMNS", -1);
-        for (j = 0; j < ncolumns; j++) {
-          grn_id range_id;
-          grn_output_array_open(ctx, outbuf, output_type, "COLUMN", -1);
-          GRN_BULK_REWIND(&buf);
-          grn_column_name_(ctx, columns[j], &buf);
-          grn_output_obj(ctx, outbuf, output_type, &buf, NULL);
-          /* column range */
-          range_id = grn_obj_get_range(ctx, columns[j]);
-          if (range_id == GRN_ID_NIL) {
-            GRN_TEXT_PUTS(ctx, outbuf, "null");
-          } else {
-            int name_len;
-            grn_obj *range_obj;
-            char name_buf[GRN_TABLE_MAX_KEY_SIZE];
-
-            range_obj = grn_ctx_at(ctx, range_id);
-            name_len = grn_obj_name(ctx, range_obj, name_buf,
-                                    GRN_TABLE_MAX_KEY_SIZE);
-            GRN_BULK_REWIND(&buf);
-            GRN_TEXT_PUT(ctx, &buf, name_buf, name_len);
-            grn_output_obj(ctx, outbuf, output_type, &buf, NULL);
-          }
-          grn_output_array_close(ctx, outbuf, output_type);
-        }
-        grn_output_array_close(ctx, outbuf, output_type);
-      }
-      for (i = 0;; i++) {
-        grn_output_array_open(ctx, outbuf, output_type, "HITS", -1);
-        for (j = 0; j < ncolumns; j++) {
-          GRN_BULK_REWIND(&buf);
-          grn_obj_get_value(ctx, columns[j], *v, &buf);
-          grn_output_obj(ctx, outbuf, output_type, &buf, NULL);
-        }
-        grn_output_array_close(ctx, outbuf, output_type);
-        v++;
-        if (v < ve) {
-
-        } else {
-          break;
-        }
-      }
+    if (GRN_BULK_VSIZE(&(format->columns)) > 0) {
+      output_result_set = GRN_TRUE;
     }
-    grn_output_array_close(ctx, outbuf, output_type);
-  } else {
-    grn_obj *range = grn_ctx_at(ctx, uvector->header.domain);
-    if (range && range->header.type == GRN_TYPE) {
-      int value_size = ((struct _grn_type *)range)->obj.range;
-      char *v = (char *)GRN_BULK_HEAD(uvector),
-        *ve = (char *)GRN_BULK_CURR(uvector);
-      grn_output_array_open(ctx, outbuf, output_type, "VECTOR", -1);
-      if (v < ve) {
-        for (;;) {
-          grn_obj value;
-          GRN_OBJ_INIT(&value, GRN_BULK, 0, uvector->header.domain);
-          grn_bulk_write_from(ctx, &value, v, 0, value_size);
-          grn_output_obj(ctx, outbuf, output_type, &value, NULL);
-
-          v += value_size;
-          if (v < ve) {
-
-          } else {
-            break;
-          }
-        }
-      }
-      grn_output_array_close(ctx, outbuf, output_type);
-    } else {
-      grn_id *v = (grn_id *)GRN_BULK_HEAD(uvector),
-        *ve = (grn_id *)GRN_BULK_CURR(uvector);
-      grn_output_array_open(ctx, outbuf, output_type, "VECTOR", ve - v);
-      if (v < ve) {
-        grn_obj key;
-        GRN_OBJ_INIT(&key, GRN_BULK, 0, range->header.domain);
-        for (;;) {
-          if (range->header.type != GRN_TABLE_NO_KEY) {
-            grn_table_get_key2(ctx, range, *v, &key);
-            grn_output_obj(ctx, outbuf, output_type, &key, NULL);
-            GRN_BULK_REWIND(&key);
-          } else {
-            grn_obj id;
-            GRN_UINT32_INIT(&id, 0);
-            GRN_UINT32_SET(ctx, &id, *v);
-            grn_output_obj(ctx, outbuf, output_type, &id, NULL);
-            GRN_OBJ_FIN(ctx, &id);
-          }
-          v++;
-          if (v < ve) {
-
-          } else {
-            break;
-          }
-        }
-        GRN_OBJ_FIN(ctx, &key);
-      }
-      grn_output_array_close(ctx, outbuf, output_type);
+    if (format->flags & GRN_OBJ_FORMAT_WITH_WEIGHT) {
+      with_weight = GRN_TRUE;
     }
   }
-  GRN_OBJ_FIN(ctx, &buf);
+
+  if (output_result_set) {
+    grn_output_uvector_result_set(ctx, outbuf, output_type, uvector, format);
+    return;
+  }
+
+  range = grn_ctx_at(ctx, uvector->header.domain);
+  range_is_type = (range->header.type == GRN_TYPE);
+  if (range_is_type) {
+    unsigned int i, n;
+    char *raw_elements;
+    unsigned int element_size;
+    grn_obj element;
+
+    raw_elements = GRN_BULK_HEAD(uvector);
+    element_size = GRN_TYPE_SIZE(DB_OBJ(range));
+    n = GRN_BULK_VSIZE(uvector) / element_size;
+
+    grn_output_array_open(ctx, outbuf, output_type, "VECTOR", n);
+    GRN_OBJ_INIT(&element, GRN_BULK, 0, uvector->header.domain);
+    for (i = 0; i < n; i++) {
+      GRN_BULK_REWIND(&element);
+      grn_bulk_write_from(ctx, &element, raw_elements + (element_size * i),
+                          0, element_size);
+      grn_output_obj(ctx, outbuf, output_type, &element, NULL);
+    }
+    GRN_OBJ_FIN(ctx, &element);
+    grn_output_array_close(ctx, outbuf, output_type);
+  } else {
+    unsigned int i, n;
+    grn_obj id_value;
+    grn_obj key_value;
+
+    GRN_UINT32_INIT(&id_value, 0);
+    GRN_OBJ_INIT(&key_value, GRN_BULK, 0, range->header.domain);
+
+    n = grn_vector_size(ctx, uvector);
+    if (with_weight) {
+      grn_output_map_open(ctx, outbuf, output_type, "WEIGHT_VECTOR", n);
+    } else {
+      grn_output_array_open(ctx, outbuf, output_type, "VECTOR", n);
+    }
+
+    for (i = 0; i < n; i++) {
+      grn_id id;
+      unsigned int weight;
+
+      id = grn_uvector_get_element(ctx, uvector, i, &weight);
+      if (range->header.type == GRN_TABLE_NO_KEY) {
+        GRN_UINT32_SET(ctx, &id_value, id);
+        grn_output_obj(ctx, outbuf, output_type, &id_value, NULL);
+      } else {
+        GRN_BULK_REWIND(&key_value);
+        grn_table_get_key2(ctx, range, id, &key_value);
+        grn_output_obj(ctx, outbuf, output_type, &key_value, NULL);
+      }
+
+      if (with_weight) {
+        grn_output_uint64(ctx, outbuf, output_type, weight);
+      }
+    }
+
+    if (with_weight) {
+      grn_output_map_close(ctx, outbuf, output_type);
+    } else {
+      grn_output_array_close(ctx, outbuf, output_type);
+    }
+
+    GRN_OBJ_FIN(ctx, &id_value);
+    GRN_OBJ_FIN(ctx, &key_value);
+  }
+  grn_obj_unlink(ctx, range);
 }
 
 static inline void
