@@ -1087,7 +1087,7 @@ grn_pat_lcp_search(grn_ctx *ctx, grn_pat *pat, const void *key, uint32_t key_siz
 }
 
 inline static grn_rc
-__grn_pat_del(grn_ctx *ctx, grn_pat *pat, const char *key, uint32_t key_size, int shared,
+_grn_pat_del(grn_ctx *ctx, grn_pat *pat, const char *key, uint32_t key_size, int shared,
              grn_table_delete_optarg *optarg)
 {
   grn_pat_delinfo *di;
@@ -1226,7 +1226,7 @@ _grn_pat_delete(grn_ctx *ctx, grn_pat *pat, const void *key, uint32_t key_size,
     }
     return GRN_INVALID_ARGUMENT;
   }
-  return __grn_pat_del(ctx, pat, key, key_size, 0, optarg);
+  return _grn_pat_del(ctx, pat, key, key_size, 0, optarg);
 }
 
 grn_rc
@@ -1457,7 +1457,7 @@ grn_pat_delete_with_sis(grn_ctx *ctx, grn_pat *pat, grn_id id,
       shared = 0;
     }
     key_size = PAT_LEN(rn);
-    if (key && key_size) { __grn_pat_del(ctx, pat, key, key_size, shared, NULL); }
+    if (key && key_size) { _grn_pat_del(ctx, pat, key, key_size, shared, NULL); }
     if (si) {
       grn_id *p, sid;
       uint32_t lkey = 0;
@@ -2777,12 +2777,13 @@ rk_emit(rk_tree_node *rn, char **str)
 } while (0)
 
 static uint32_t
-rk_conv(const char *str, uint32_t str_len, char *buf, uint32_t buf_size, uint8_t *statep)
+rk_conv(const char *str, uint32_t str_len, uint8_t *buf, uint32_t buf_size, uint8_t *statep)
 {
   uint32_t l;
   uint8_t state = 0;
   rk_tree_node *rn;
-  char *e, *oc = buf, *oe = oc + buf_size;
+  char *e;
+  uint8_t *oc = buf, *oe = oc + buf_size;
   const uint8_t *ic = (uint8_t *)str, *ic_ = ic, *ie = ic + str_len;
   while (ic < ie) {
     if ((rn = rk_lookup(state, *ic))) {
